import pandas as pd
import torch
from sklearn.model_selection import train_test_split
from torch import nn
import torch.optim as optim
import time
import numpy as np
import matplotlib.pyplot as plt

def load_file(file_name):
    """
    Loads file file_name. 
    Assumes it uses spaces for separators.
    Returns a PyTorch tensor with float32 type values.
    """

    # Step 1: Load data from a file
    data = pd.read_csv(file_name, sep='\\s+', header=None)
    
    # Step 2: Convert to NumPy array
    data_array = data.to_numpy()

    # Step 3: Convert to PyTorch tensor
    data_tensor = torch.tensor(data_array, dtype=torch.float32)
    
    return data_tensor
    
    
def sinecosineone(dttmval,period):
    """
    Use both sine and cosine for each of the periodic
    date and time values like hour of day or
    day of month
    
    """
    # Convert dttmval to radians
    radians = (2 * np.pi * dttmval) / period
    # Apply sine and cosine transformations
    sin_dttmval = np.sin(radians)
    cos_dttmval = np.cos(radians)
    return sin_dttmval, cos_dttmval
    
def sinecosineall(X):
    """
    Column         Number
    
    day_of_week      7
    day_of_month     31
    day_of_year      366
    hour_of_day      24
    month            12
    quarter          4
    week_of_year     52
    week_of_month    5
    
    """
    
    new_X = []
    
    for row in X:
        day_of_week_sine, day_of_week_cosine = sinecosineone(row[0],7)
        day_of_month_sine, day_of_month_cosine = sinecosineone(row[1],31)
        day_of_year_sine, day_of_year_cosine = sinecosineone(row[2],366)
        hour_of_day_sine, hour_of_day_cosine = sinecosineone(row[3],24)
        month_sine, month_cosine = sinecosineone(row[4],12)
        quarter_sine, quarter_cosine = sinecosineone(row[5],4)
        week_of_year_sine, week_of_year_cosine = sinecosineone(row[6],52)
        week_of_month_sine, week_of_month_cosine = sinecosineone(row[7],5)
        
        new_X.append([day_of_week_sine,day_of_week_cosine,day_of_month_sine,day_of_month_cosine,
        day_of_year_sine, day_of_year_cosine,hour_of_day_sine, hour_of_day_cosine,
        month_sine,month_cosine,quarter_sine,quarter_cosine,
        week_of_year_sine,week_of_year_cosine,week_of_month_sine,week_of_month_cosine])

    return torch.tensor(new_X)
    
def get_train_test(data_tensor):
    """
    
    Returns X and y training and testing tensors
    
    """

    # Assume data_tensor is your full dataset as a PyTorch tensor
    # Splitting data into features (X) and target (y) if applicable
    # Example assumes the last column is the target variable
    # Modified to skip year. All the data is from the same year.
    
    X = data_tensor[:, :-2]
    y = data_tensor[:, -1]
    
    new_X = sinecosineall(X)
    
    # Use train_test_split from scikit-learn
    X_train, X_test, y_train, y_test = train_test_split(new_X, y, test_size=0.2, random_state=42)
    
    # gets a second size 1 column
    
    y_train_tensor = y_train.unsqueeze(1)

    y_test_tensor = y_test.unsqueeze(1)
        
    return X_train,y_train_tensor,X_test,y_test_tensor

def batchify_data(x_data, y_data, batch_size):
    """Takes a set of data points and labels and groups them into batches."""
    # Only take batch_size chunks (i.e. drop the remainder)
    N = int(len(x_data) / batch_size) * batch_size
    batches = []
    for i in range(0, N, batch_size):
        batches.append({
            'x': x_data[i:i + batch_size],
            'y': y_data[i:i + batch_size]
        })
    return batches
    
# Step 1: Define the model
class SimpleRegressionModel(nn.Module):
    def __init__(self):
        super(SimpleRegressionModel, self).__init__()
        self.fc1 = nn.Linear(16, 64)  # First fully connected layer
        self.fc2 = nn.Linear(64, 32) # Second fully connected layer
        self.fc3 = nn.Linear(32, 1)  # Output layer

    def forward(self, x):
        x = torch.relu(self.fc1(x))  # ReLU activation for non-linearity
        x = torch.relu(self.fc2(x))
        x = self.fc3(x)  # Output layer (no activation for regression)
        return x

def train(train_batches,learningrate,num_epochs):
    """ 
    Train the model
    """

    # Step 2: Initialize the model, loss function, and optimizer
    model = SimpleRegressionModel()
    criterion = nn.MSELoss()  # Mean Squared Error loss
    #optimizer = optim.Adam(model.parameters(), lr=learningrate)
    optimizer = optim.SGD(model.parameters(), lr=learningrate)

    # Step 3: Training loop
    for epoch in range(num_epochs):
        model.train()  # Set the model to training mode
    
        for batch in train_batches:
            X_train_batch = batch['x']
            y_train_batch = batch['y']
    
            # Zero the gradient buffers
            optimizer.zero_grad()
    
            # Forward pass
            outputs = model(X_train_batch)
    
            loss = criterion(outputs, y_train_batch)
      
            # Backward pass and optimize
            loss.backward()
            optimizer.step()
    
        if (epoch+1) % 10 == 0:  # Print every 10 epochs
            print(f'Epoch [{epoch+1}/{num_epochs}], Loss: {loss.item():.4f}')
            
    return model,criterion

def testmodel(model,criterion,X_test_tensor,y_test_tensor):
    """
    Test model
    """
    
    # Step 4: Evaluate the model (optional)
    model.eval()  # Set the model to evaluation mode
    with torch.no_grad():
        predictions = model(X_test_tensor)
        test_loss = criterion(predictions, y_test_tensor)
        print(f'Test Loss: {test_loss.item():.4f}')
        
    return predictions
    
def print_compare(predictions,y_test_tensor):
    """
    print a simple comparison of predicted cpu use
    versus actual
    """
    
    length = predictions.shape[0]
    
    total_abs_diff = 0.0
    max_abs_diff = 0.0
    print('cpu percent diff 10 or higher:')
    for i in range(length):
        abs_diff = abs(float(predictions[i][0])-float(y_test_tensor[i][0]))
        total_abs_diff += abs_diff
        if abs_diff > max_abs_diff:
            max_abs_diff = abs_diff
        if abs_diff >= 10.0:
            print(str(round(abs_diff,2))+'%')
        
    print('average abs(Predicted CPU% - actual CPU%) = '+str(round(total_abs_diff/length,2))+'%')
    print('maximum abs(Predicted CPU% - actual CPU%) = '+str(round(max_abs_diff,2))+'%')
       
if __name__ == '__main__':

    # Capture time so I can keep track of how long it takes to run this script
    
    before = time.perf_counter()
    
    # adjustable hyperparameters
    
    learningrate = 0.001 # learning rate for model training
    
    num_epochs = 200  # Number of training epochs

    batch_size = 2   # group training data into this many in a batch
    
    file_name = 'testdata.txt'

    # Load space separated file as data tensor float32 values

    data_tensor = load_file(file_name)
    
    # split into training and testing data

    X_train_tensor,y_train_tensor,X_test_tensor,y_test_tensor = get_train_test(data_tensor)
    
    # group into batches
    
    train_batches = batchify_data(X_train_tensor, y_train_tensor, batch_size)
    
    # train model
    
    model,criterion = train(train_batches,learningrate,num_epochs)
    
    # test model
    
    predictions = testmodel(model,criterion,X_test_tensor,y_test_tensor)
    
    # print predicted versus actual percentage off
    
    print_compare(predictions,y_test_tensor)

    # Finish run time calculation and print out
    
    after = time.perf_counter()
    
    print(' ')  
    print('Total run time in seconds: {0:<10.0f}'.format(after - before))
    
    # graph all data versus predicted Y
    
    X = data_tensor[:, :-2]
    new_X = sinecosineall(X)
    with torch.no_grad():
        predictions = model(new_X)
    #actual
    plt.plot(data_tensor[:,9],'b',label='actual')
    #predicted
    plt.plot(predictions,'r',label='predicted')
    plt.ylabel('CPU Percent Used')
    plt.legend()
    plt.show()
    